import numpy as np
from abc import ABC, abstractmethod
from typing import Dict

class Layer(ABC):
    """
    Base layer class.

    All layers must inherit from this class
    because the NN works with Layer objects
    """

    def __init__(self) -> None:
        self.forward_inputs = None
        self.forward_outputs = None
        self.backward_inputs = None
        self.backward_outputs = None
        self._freeze = False
        self._train = False

    def __repr__(self, extra_fields = {}):
        return str({'Type:': type(self).__name__, 
                'forward_inputs': self.forward_inputs,
                'forward_outputs': self.forward_outputs,
                'backward_inputs': self.backward_inputs,
                'backward_outputs': self.backward_outputs, 
                **extra_fields})
    
    @abstractmethod
    def forward(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def backward(self, output_gradient: float) -> np.ndarray:
        pass

    def get_trainable_params(self) -> Dict[str, np.ndarray]:
        """Returns a dictionary of all trainable parameters by the layer"""
        return {}

    def get_trainable_params_gradients(self) -> Dict[str, np.ndarray]:
        """
        Returns a dictionary of all trainable parameters GRADIENTS by the layer
        This is mainly used by the optimizer
        """
        return {}

    def update_trainable_params(self, optimized_params: Dict[str, np.ndarray]) -> None:
        """Update trainable params. Order of optimized_params is the same as get_trainable_params_gradients"""
        pass

    def get_train(self):
        return self._train

    def set_train(self, state):
        assert type(state) == bool
        self._train = state
    
    def freeze(self):
        """
        Freeze layer. Useful when doing transfer learning (pending implementation)
        """
        self._freeze = True
    
    def unfreeze(self):
        self._freeze = False
    
class Dense(Layer):
    def __init__(self, n_rows: int = None, n_columns: int = None, weight_init_scale = 0.1, seed: int = None) -> None:
        if seed:
            np.random.seed(seed)
        self.weights = np.random.randn(n_rows, n_columns) * weight_init_scale
        self.bias = np.random.randn(n_rows, 1)
        self.bias_prime = np.zeros(self.bias.shape)
        self.weights_prime = np.zeros(self.weights.shape)
        super().__init__()

    def __repr__(self):
        fields = {'weights': self.weights, 'bias': self.bias, 
                'weights_prime': self.weights_prime, 'bias_prime': self.bias_prime}
        return super().__repr__(extra_fields = fields)

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.forward_inputs = x
        self.forward_outputs = np.dot(self.weights, self.forward_inputs) + self.bias
        return self.forward_outputs

    def backward(self, output_gradient: float) -> np.ndarray:
        self.backward_inputs = output_gradient
        self.backward_outputs = np.dot(self.weights.T, self.backward_inputs)
        # compute gradients
        self.weights_prime = np.dot(self.backward_inputs, self.forward_inputs.T)
        self.bias_prime = np.sum(self.backward_inputs, axis=1, keepdims=True)
        return self.backward_outputs

    def get_trainable_params(self) -> Dict[str, np.ndarray]:
        return {'weights': self.weights, 'bias': self.bias}

    def get_trainable_params_gradients(self) -> Dict[str, np.ndarray]:
        return {'weights_prime': self.weights_prime, 'bias_prime': self.bias_prime}

    def update_trainable_params(self, optimized_params: Dict[str, np.ndarray]) -> None:
        if not self._freeze:
            # update weights and biases
            self.weights -= optimized_params['weights_prime']
            self.bias -= optimized_params['bias_prime']
        return None

class Dropout(Layer):
    def __init__(self, drop_prob: float = 0) -> None:
        assert 0 <= drop_prob < 1, "Drop probability rate should be between 0 and 0.9"
        self.drop_prob = drop_prob
        self.keep_prob = 1 - self.drop_prob
        super().__init__()

    def forward(self, x: np.ndarray) -> np.ndarray:
        self.forward_inputs = x
        matrix_probs = [self.drop_prob, 1 - self.drop_prob]
        if self._train:
            matrix_probs = [self.drop_prob, 1 - self.drop_prob]
            self.drop_matrix = np.random.choice([0, 1], size=x.shape, p=matrix_probs)
            self.forward_outputs = np.multiply(self.drop_matrix, self.forward_inputs) / self.keep_prob
        else:
            self.forward_outputs = self.forward_inputs
        return self.forward_outputs

    def backward(self, output_gradient: float) -> np.ndarray:
        self.backward_inputs = output_gradient
        self.backward_outputs = np.multiply(self.drop_matrix, output_gradient) / self.keep_prob
        return self.backward_outputs

####################
# Activation Layer #
####################

class ActivationLayer(Layer):
    def __init__(self, activation, activation_prime) -> None:
        self.activation = activation
        self.activation_prime = activation_prime
        super().__init__()

    def forward(self, x) -> np.ndarray:
        self.forward_inputs = x
        self.forward_outputs = self.activation(self.forward_inputs)
        return self.forward_outputs

    def backward(self, output_gradient: float) -> np.ndarray:
        self.backward_inputs = output_gradient
        self.backward_outputs = np.multiply(self.backward_inputs, self.activation_prime(self.forward_inputs))
        return self.backward_outputs

class Sigmoid(ActivationLayer):
    def __init__(self) -> None:
        sigmoid = lambda x: 1 / (1 + np.exp(-x))
        sigmoid_prime = lambda x: sigmoid(x) * (1 - sigmoid(x))
        super().__init__(sigmoid, sigmoid_prime)

class ReLU(ActivationLayer):
    def __init__(self) -> None:
        relu = lambda x: np.maximum(0, x)
        relu_prime = lambda x: (x>0).astype(x.dtype)
        super().__init__(relu, relu_prime)

# softmax is a bit different from other activations,
# its more like a fully connected since it depends on all
# the input variables to compute the sum to create the
# probability distribution which adds up to 1.
# thats why it receives a different treatment and inherits
# directoly from the Layer superclass
class Softmax(Layer):
    def forward(self, x) -> np.ndarray:
        self.forward_inputs = x
        input_stable = np.exp(x - np.max(x, axis=0))
        self.forward_outputs =  input_stable / input_stable.sum(axis=0)
        return self.forward_outputs
    
    def backward(self, output_gradient: float) -> np.ndarray:
        self.backward_inputs = output_gradient
        tmp = np.sum(self.backward_inputs * self.forward_outputs, axis = 0)
        self.backward_outputs = self.forward_outputs * (self.backward_inputs - tmp)
        return self.backward_outputs
        


import numpy as np
import dill
from redmind.network import NeuralNetwork

def one_hot_encode(position: int, num_classes: int):
    """
    One hot encodes a given class into a (1, num_classes) row vector
    """
    zero_vector = np.zeros((1, num_classes), dtype=int) 
    zero_vector[0][position] = 1
    # return 1d row vector
    return zero_vector.reshape(-1)

def save_model(nn: NeuralNetwork, filename: str = 'nn.dill') -> None:
    if isinstance(nn, NeuralNetwork):
        print(f"Saving Neural Network into {filename}")
        with open(filename, 'wb') as f:
            dill.dump(nn, f)
    else:
        print("Please provide a valid Neural Network")
    return None

def load_model(filename: str = 'nn.dill') -> NeuralNetwork:
    print(f"Loading NN {filename}")
    with open(filename, 'rb') as f:
        nn = dill.load(f)
    return nn


import numpy as np
from typing import List
from redmind.layers import Layer

class NeuralNetwork:
    def __init__(self, layers: List[Layer], verbose=False) -> None:
        self.layers = layers
        self._verbose = verbose
        if self._verbose:
            print(f"Neural Network initialized with {len(self.layers)} layers")

    def __repr__(self):
        return f"NeuralNetwork(layers={self.layers})"

    def forward(self, x: np.ndarray = None) -> np.ndarray:
        out = x
        for layer in self.layers:
            out = layer.forward(out)
        return out

    def backward(self, gradient: float = None) -> None:
        grad = gradient
        for layer in reversed(self.layers):
            grad = layer.backward(output_gradient=grad)
        return None

    def predict(self, x: np.ndarray = None) -> np.ndarray:
        """
        Same as forward, its just used for convention
        """
        return self.forward(x)

    def set_train(self, state=False):
        """
        Set network train state
        This is useful if you are using dropout layers. If so make sure properly
        changing the training states for better accuracy
        """
        if self._verbose:
            print(f"updating NN layers training to: {state}")
        for layer in self.layers:
            layer.set_train(state=state)

    def set_verbose(self, state):
        assert type(state) == bool
        self._verbose = state





import numpy as np

########################
# Regression functions #
########################
def mse(y, y_pred) -> np.float64:
    squared_error = np.power(y_pred - y, 2)
    return np.mean(squared_error)
    
def mse_prime(y, y_pred) -> np.ndarray:
    return 2 * (y_pred - y)

#########################
# Categorical functions #
#########################

##
# Multi class classification [0 0 1 0 0]
# or one hot encoded vector
##
def cross_entropy(y, y_pred):
    return -np.mean(y * np.log(y_pred))

def cross_entropy_prime(y, y_pred):
    return y_pred - y

##
# two-class / binary classification  [0 1] [0] [1]
##
def binary_cross_entropy(y ,y_pred) -> np.float64:
    return  -np.mean((y * np.log(y_pred) + (1 - y) * np.log(1 - y_pred)))

def binary_cross_entropy_prime(y, y_pred) -> np.ndarray:
    return -((y_pred-y) / ((y_pred - 1) * y_pred))

#################################
# Learning rate decay functions #
#################################
def lr_decay(learning_rate, epoch, decay_rate):
    """Standard learning rate decay algorithm"""
    return (1 / (1 + decay_rate * epoch)) * learning_rate


"""
Neural network optimizers
"""
import numpy as np
from abc import ABC, abstractmethod
from redmind.layers import Layer
from redmind.network import NeuralNetwork

def init_velocity_vector(layers):
    velocity = {}
    # build layers velocity dict with np zeros array for each trainable pram
    for idx, layer in enumerate(layers):
        trainable_params = layer.get_trainable_params_gradients()
        velocity[idx] = trainable_params
        for param, grads in trainable_params.items():
            velocity[idx][param] = np.zeros(grads.shape)
    return velocity

class Optimizer(ABC):
    """
    Only one optimizer can be assigned to the entire NN

    The optimizer is in charge of keeping track of optimization
    variable states for all layers. Check Adam optimizer for reference.

    Optimizer workflow
    1. Loop through self.layers (coming from Neural Network)
    2. Fetch all trainable parameters
    3. Optimize those parameters according to strategy
    4. (optional) track optimization
    5. Update layer parameters on learning rate scale
    """
    def __init__(self, network: NeuralNetwork):
        assert isinstance(network, NeuralNetwork), "network should be a NeuralNetwork object"
        self.layers = network.layers
    
    def set_learning_rate(self, learning_rate: float) -> None:
        self.learning_rate = learning_rate

    @abstractmethod
    def __call__(self) -> None:
        """Runs optimizer for all trainable parameters returned by the layer"""
        
class GradientDescent(Optimizer):
    def __call__(self) -> None:
        for layer in self.layers:
            trainable_params = layer.get_trainable_params_gradients()
            for param, grads in trainable_params.items():
                trainable_params[param] = grads * self.learning_rate
            layer.update_trainable_params(trainable_params)

class Momentum(Optimizer):
    beta = 0.9

    def __init__(self, network: NeuralNetwork):
        super().__init__(network)
        self.gradients_velocity = init_velocity_vector(self.layers)


    def __call__(self) -> None:
        for idx, layer in enumerate(self.layers):
            trainable_params = layer.get_trainable_params_gradients()
            for param, grads in trainable_params.items():
                self.gradients_velocity[idx][param] = self.beta * self.gradients_velocity[idx][param] + (1 - self.beta) * grads
                trainable_params[param] = self.gradients_velocity[idx][param] * self.learning_rate
            layer.update_trainable_params(trainable_params)

class RMSprop(Optimizer):
    beta = 0.9
    epsilon = 1e-7

    def __init__(self, network: NeuralNetwork):
        super().__init__(network)
        self.gradients_velocity = init_velocity_vector(self.layers)

    def __call__(self) -> None:
        for idx, layer in enumerate(self.layers):
            trainable_params = layer.get_trainable_params_gradients()
            for param, grads in trainable_params.items():
                self.gradients_velocity[idx][param] = self.beta * self.gradients_velocity[idx][param] + (1 - self.beta) * np.power(grads, 2)
                trainable_params[param] = (grads / np.sqrt(self.gradients_velocity[idx][param] + self.epsilon)) * self.learning_rate
            layer.update_trainable_params(trainable_params)

class Adam(Optimizer):
    """Adam is a combination of momentum and RMSprop, thats why the velocity names"""
    beta1 = 0.9
    beta2 = 0.999
    epsilon = 1e-7

    def __init__(self, network: NeuralNetwork):
        super().__init__(network)
        self.momentum_velocity = init_velocity_vector(self.layers)
        self.rmsprop_velocity = init_velocity_vector(self.layers)

    def __call__(self) -> None:
        for idx, layer in enumerate(self.layers):
            trainable_params = layer.get_trainable_params_gradients()
            for param, grads in trainable_params.items():
                self.momentum_velocity[idx][param] = ((self.beta1 * self.momentum_velocity[idx][param]) + ((1 - self.beta1) * grads)) 
                self.rmsprop_velocity[idx][param] = ((self.beta2 * self.rmsprop_velocity[idx][param]) + ((1 - self.beta2) * np.power(grads, 2)))
                trainable_params[param] = (self.momentum_velocity[idx][param] / np.sqrt(self.rmsprop_velocity[idx][param] + self.epsilon)) * self.learning_rate
            layer.update_trainable_params(trainable_params)




import numpy as np
import matplotlib.pyplot as plt
import redmind.functions as fn
from typing import List, Dict
from redmind.network import NeuralNetwork
from redmind.layers import Layer
from redmind.dataloader import Dataloader
from redmind.optimizers import Optimizer
from redmind.optimizers import GradientDescent


class Trainer():
    """
    Trainer class makes training easier. It has predefined trainig functions so you
    only need to send the arguments. You can also train manually if you want.
    """
    def __init__(self, network: NeuralNetwork, learning_rate: float = 1e-2, lr_decay_function = None, decay_rate: float = None, 
                cost_function = fn.mse, grad_function = fn.mse_prime, optimizer: Optimizer = None):
        assert isinstance(network, NeuralNetwork), "network should be a NeuralNetwork object"
        self.network = network
        if not optimizer:
            optimizer = GradientDescent(self.network)
        assert isinstance(optimizer, Optimizer), "optimizer should be an Optimizer object"
        self.learning_rate = learning_rate
        self.lr_decay_function = lr_decay_function
        self.decay_rate = decay_rate
        self.optimizer = optimizer
        self.costs = {}
        self.cost_function = cost_function
        self.grad_function = grad_function
    
    def train(self, X: np.ndarray = None, Y: np.ndarray = None, epochs: int = 100, batch_size: int = 1, early_stoping: float = 0.0):
        data = Dataloader(X=X, Y=Y, batch_size=batch_size)
        self.network.set_train(state=True)
        learning_rate = self.learning_rate
        for epoch in range(epochs):
            # learning rate decay
            if self.lr_decay_function:
                learning_rate = self.lr_decay_function(self.learning_rate, epoch, self.decay_rate)

            # Train network
            self.optimizer.set_learning_rate(learning_rate)
            for x, y in data:
                # forward
                y_pred = self.network.forward(x)
                # calculate error and cost
                cost = self.cost_function(y, y_pred)
                self.costs[epoch] = cost
                error_gradient = self.grad_function(y, y_pred)
                # backward
                self.network.backward(gradient=error_gradient)
                # Optimize layers params
                self.optimizer()

            # print cost to console
            accuracy = round(100 - (self.costs[epoch] * 100), 3)
            print(f"epoch: {epoch + 1}/{epochs}, cost: {round(self.costs[epoch], 4)}, accuracy: {accuracy}%")

            # Early stoping
            if early_stoping > 0.0 and accuracy >= early_stoping:
                print(f"Early stoping threshold reached (over {early_stoping}%)")
                break
        self.network.set_train(state=False)

    def graph_costs(self) -> None:
        plt.plot(list(self.costs.keys()), list(self.costs.values()))
        plt.xlabel("Epoch")
        plt.ylabel("Cost")
        plt.show()

    def grad_check(self, X: np.ndarray = None, Y: np.ndarray = None, epsilon=1e-7):
        """
        Performs gradient checking on the given NN
        It is recommended to input X and Y with few elements since grad check differences
        can add up and give you high difference margin.
        """
        # Helper functions for grad check
        def params_to_vector(params: Dict[int, np.ndarray]) -> np.ndarray:
            """Converts layers parameters to a row vector"""
            flattened_params = []
            for layer in params.values():
                for param in layer.values():
                    flattened_params.append(param.ravel())
            return np.concatenate(flattened_params)

        def vector_to_params(vector: np.ndarray, layer_params: Dict[int, np.ndarray]) -> Dict[int, np.ndarray]:
            """Returns dict of params with values from row vector"""
            assert params_to_vector(layer_params).size == vector.size, "vector and layer_params sizes do not match"
            updated_params = layer_params.copy()
            for k, v in layer_params.items():
                for param_name, val in v.items():
                    # get number of elements in the parameter values
                    param_size = val.size
                    # get items from idx 0 to param_size from vector
                    new_values = vector[:param_size]
                    assert val.size == new_values.size, "values sizes do not match"
                    # set values for that param in that layer in updated_params
                    updated_params[k][param_name] = new_values.reshape(updated_params[k][param_name].shape)
                    # remove assigned values from vector
                    vector = np.delete(vector, range(param_size))
            assert vector.size == 0, "Something is wrong, vector still has values!"
            return updated_params

        def update_layer_params(layer_params: Dict[int, np.ndarray]) -> None:
            """
            Update network layer parameters with given layer_params values.
            Order of the dict should be the same as self.network.layers
            """
            for idx, layer in enumerate(self.network.layers):
                for k, v in layer_params[idx].items():
                    layer.__dict__[k] = v
        
        print("Starting gradient checking")
        # compute NN gradients through forward and backward pass
        y_pred = self.network.forward(X)
        error_gradient = self.grad_function(Y, y_pred)
        self.network.backward(gradient=error_gradient)
        ###
        # get gradients for each layer and flatten in a row vector
        ###
        gradients = {}
        for idx, layer in enumerate(self.network.layers):
            gradients[idx] = layer.get_trainable_params_gradients()
        
        # If you are using mini batches then you need to divide the gradients by the number of samples
        # This is not proven, but it worked for me, I might need to research further
        flattened_gradients = params_to_vector(gradients)
        ###
        # get original params for each layer and flatten in single row vector
        ###
        original_params = {}
        for idx, layer in enumerate(self.network.layers):
            original_params[idx] = layer.get_trainable_params()

        flattened_params = params_to_vector(original_params)
        ###
        # Compute numgrad with cost function for each param
        ###
        numgrad = np.zeros(flattened_params.shape)
        preturb = np.zeros(numgrad.shape)

        for i in range(numgrad.size):
            preturb[i] = epsilon
            # convert params vector plus epsilon
            plus_params = vector_to_params(vector = flattened_params + preturb, layer_params = original_params)
            update_layer_params(plus_params)
            y_plus = self.network.forward(X)
            loss_plus = self.cost_function(Y, y_plus)

            # convert params vector minus epsilon
            minus_params = vector_to_params(vector = flattened_params - preturb, layer_params = original_params)
            update_layer_params(minus_params)
            y_minus = self.network.forward(X)
            loss_minus = self.cost_function(Y, y_minus)
            # compute the gradient
            numgrad[i] = (loss_plus - loss_minus) / (2*epsilon)
            # reset preturb
            preturb[i] = 0

        numerator = np.linalg.norm(numgrad - flattened_gradients)
        denominator = np.linalg.norm(flattened_gradients) + np.linalg.norm(numgrad) 
        difference = numerator / denominator
        if difference > 2e-7 or np.isnan(difference):
            print ("\033[93m" + "There is a mistake in the backward propagation! difference = " + str(difference) + "\033[0m")
        else:
            print ("\033[92m" + "Your backward propagation works perfectly fine! difference = " + str(difference) + "\033[0m")